------------------------------------------------------------------------------
--- This module defines the data structure for GNN trees
--- (GNN stands for Generalizing Narrowing Trees) 
--- and some associated functions
--- 
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module GNNTrees(GNNTree(..),mark_memo) where

import FlatCurry

data GNNTree = Leaf   Expr             --standard leaf
             | LeafM  Expr Expr        --leaf with memo expression (for renaming)  
             | TreeN  Expr [GNNTree]   --needed narrowing step
             | TreeNM Expr [GNNTree]   --needed narrowing step (with a call back)
             | TreeD  Expr [GNNTree]   --deconstruction step
             | TreeG  Expr [GNNTree]   --generalization step

--marks nodes in a GNN tree which have been used for memo steps:
mark_memo :: GNNTree -> GNNTree
mark_memo t = mark (collectMemos t) t

--marking of nodes:
mark :: [Expr] -> GNNTree -> GNNTree
mark es t = case t of
              Leaf l     -> Leaf l
              LeafM l e  -> LeafM l e
              TreeG e ts -> TreeG e (map (mark es) ts)
              TreeD e ts -> TreeD e (map (mark es) ts)
              TreeN e ts -> if e `elem` es then TreeNM e (map (mark es) ts)
                                            else TreeN e (map (mark es) ts)
   

--collects the expressions in LeafM nodes:
collectMemos :: GNNTree -> [Expr]
collectMemos t = case t of
                   LeafM _ e -> [e]
                   Leaf _ -> []
                   TreeD _ es -> concat (map collectMemos es)
                   TreeN _ es -> concat (map collectMemos es)
                   TreeG _ es -> concat (map collectMemos es)
