------------------------------------------------------------------------------
--- This module contains functions for implementing generalization steps
--- (within the "generalizing needed narrowing")
--- 
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module General(hasGEN,generalize,delGEN,delGEN2) where

import FlatCurry
--import Unsafe

import Utils(maxVarIndex,triple1st,triple2nd)
--import OffPretty(ppExpr)

--checks whether there are occurrences of GEN:
hasGEN :: Expr -> Bool
hasGEN (Var _) = False
hasGEN (Lit _) = False
hasGEN (Comb ftype (_,fname) es) =
  case ftype of
        FuncCall -> if fname == "GEN" && length es == 1 
                       then True 
                       else or (map hasGEN es)
        _ -> or (map hasGEN es)
hasGEN (Free _ e) = hasGEN e
hasGEN (Or e1 e2) = (hasGEN e1) || (hasGEN e2)
hasGEN (Case _ e bs) = (hasGEN e) || (or (map hasGENbranch bs))

hasGENbranch (Branch _ e) = hasGEN e

--generalization proper:
generalize :: Int -> Expr -> [Expr]
generalize fresh e = let (ge,ges,_) = gen (fresh + (maxVarIndex e) + 1) e
               in ge : ges

--gen  varnum,expresion_a_generalizar = exprs generalizada, expresiones nuevas, contador de fresh var    

gen :: Int -> Expr -> (Expr,[Expr],Int)
gen n (Var v) = (Var v, [], n)
gen n (Lit l) = (Lit l, [], n)
gen n (Comb ftype (mod,fname) es) =
  case ftype of
    FuncCall -> if fname == "GEN" && length es == 1 
                   then let genes = genlist (n+1) es in 
                              (Var n,  
                               (map triple1st genes) ++ (concat (map triple2nd genes)), 
                               lastFresh genes (n+1))
                   else let genes = genlist n es in 
                              (Comb FuncCall (mod,fname) (map triple1st genes), 
                               concat (map triple2nd genes),
                               lastFresh genes n)
    _ -> let genes = genlist n es in  
                     (Comb ftype (mod,fname) (map triple1st genes), 
                      concat (map triple2nd genes),
                      lastFresh genes n)
gen n (Free vs e) = let (ge,ges,m) = gen n e in (Free vs ge, ges, m)  
gen n (Or e1 e2) = let (ge1,ges1,m) = gen n e1
                       (ge2,ges2,p) = gen m e2
                   in (Or ge1 ge2, ges1++ges2, p)
gen n (Case ctype e ces) = let 
                              (ge,ges,m) = gen n e
                              genes = genbranches m ces 
                           in 
                              (Case ctype ge (map triple1st genes),
                               ges ++ (concat (map triple2nd genes)),
                               lastFresh genes m)



genlist :: Int -> [Expr] -> [(Expr,[Expr],Int)]
genlist _ [] = []
genlist n (e:es) = let (ge,ges,m) = gen n e
                   in (ge,ges,m) : genlist m es


genbranches :: Int -> [BranchExpr] -> [(BranchExpr,[Expr],Int)]
genbranches _ [] = []
genbranches n (Branch p e : es) = let (ge,ges,m) = gen n e
                   in (Branch p ge, ges, m) : genbranches m es


lastFresh :: [(a,[Expr],Int)] -> Int -> Int
lastFresh [] n = n
lastFresh [(_,_,n)] _ = n
lastFresh (_:e:es) n = lastFresh (e:es) n

--deletes all GEN symbols from an expression:
delGEN :: Expr -> Expr
delGEN e = case e of
            Var v -> Var v
            Lit l -> Lit l
            Comb ConsCall c es -> Comb ConsCall c (map delGEN es)
            Comb FuncCall (mod,f) es -> if f=="GEN" then  (head es)    --(delGEN (head es))
                                                    else  Comb FuncCall (mod,f) (map delGEN es)
            Free vs ex -> Free vs (delGEN ex)
            Or e1 e2 -> Or (delGEN e1) (delGEN e2)
            Case ctype ex ces -> Case ctype (delGEN ex) (map delGENbranch ces)

delGENbranch (Branch p e) = Branch p (delGEN e)


--deletes all GEN symbols from an expression:
delGEN2 :: Expr -> Expr
delGEN2 e = case e of
            Var v -> Var v
            Lit l -> Lit l
            Comb ConsCall c es -> Comb ConsCall c (map delGEN2 es)
            Comb FuncCall (mod,f) es -> if f=="GEN" then  (delGEN2 (head es))
                                                    else  Comb FuncCall (mod,f) (map delGEN2 es)
            Free vs ex -> Free vs (delGEN2 ex)
            Or e1 e2 -> Or (delGEN2 e1) (delGEN2 e2)
            Case ctype ex ces -> Case ctype (delGEN2 ex) (map delGEN2branch ces)

delGEN2branch (Branch p e) = Branch p (delGEN2 e)
