------------------------------------------------------------------------------
--- This module contains several functions for pretty-printing
--- FlatCurry expressions
--- 
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module OffPretty(ppExpr,ppTree,ppPairs,ppFormatList) where

import FlatCurry

import GNNTrees(GNNTree(..))

-- pretty printer for expressions (first argument: indentation):
ppExpr _ (Var n) = ppVar n
ppExpr _ (Lit l) = ppLit l
ppExpr b (Comb _ (mod,cf) es) =
  --uncomment this for qualified names
  --"(" ++ mod ++ "." ++ cf ++ " " ++ ppList (ppExpr b) es ++ ")"
  "(" ++ cf ++ " " ++ ppList (ppExpr b) es ++ ")"
ppExpr b (Free xs e) =
  "(Free " ++ ppFormatList ppVar xs ++ (ppExpr b) e ++ ")"
ppExpr b (Or e1 e2) =
  "(Or " ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Case Rigid e cs) =
  "(Case " ++ ppExpr b e ++ " of\n " ++ ppList (ppCase (b+2)) cs
  ++ blanks b ++ ")"
ppExpr b (Case Flex e cs) =
  "(FCase " ++ ppExpr b e ++ " of\n " ++ ppList (ppCase (b+2)) cs
  ++ blanks b ++ ")"

ppVar i = "v" ++ show i

ppLit (Intc   i) = show i
ppLit (Floatc f) = show f
ppLit (Charc  c) = "'" ++ show (ord c) ++ "'"

ppCase b (Branch (Pattern (mod,l) xs) e) = 
                                       --uncomment this for qualified names
                                       --blanks b ++ "(" ++ mod ++ "." ++ l ++ " "
                            blanks b ++ "(" ++ l ++ " "
                                     ++ ppList ppVar xs
                                     ++ " -> " ++ ppExpr b e ++ ")\n"
ppCase b (Branch (LPattern l) e) = blanks b ++ "(" ++ ppLit l ++ " "
                                   ++ " -> " ++ ppExpr b e ++ ")\n"

blanks b = take b (repeat ' ')

-- format a finite list of elements (with a format function for list elems):
ppFormatList :: (a->String) -> [a] -> String
ppFormatList format elems = " [" ++ ppElems elems ++ "] "
 where
   ppElems [] = ""
   ppElems [x] = format x
   ppElems (x1:x2:xs) = format x1 ++ "," ++ ppElems (x2:xs)

ppList format elems = ppListElems elems
  where
    ppListElems [] = ""
    ppListElems [x] = format x
    ppListElems (x1:x2:xs) = format x1 ++ " " ++ ppListElems (x2:xs)

-----------------------

ppTree b (Leaf e) = "(Leaf " ++ ppExpr b e ++ ")"
ppTree b (LeafM e1 e2) = "(LeafM " ++ ppExpr b e1 ++ ppExpr b e2 ++ ")"
ppTree b (TreeN e es) = "(TreeN " ++ ppExpr b e ++ ppList (ppTree b) es ++ ")"
ppTree b (TreeNM e es) = "(TreeNM " ++ ppExpr b e ++ ppList (ppTree b) es ++ ")"
ppTree b (TreeD e es) = "(TreeD " ++ ppExpr b e ++ ppList (ppTree b) es ++ ")"
ppTree b (TreeG e es) = "(TreeG " ++ ppExpr b e ++ ppList (ppTree b) es ++ ")"

-----------------------

ppPairs :: [(Expr,Expr)] -> String -> String
ppPairs [] _ = ""
ppPairs ((lhs,rhs):es) sep = 
  ppExpr 0 lhs ++ sep ++ ppExpr 0 rhs ++ "\n" ++ ppPairs es sep

