------------------------------------------------------------------------------
--- This module includes functions for transforming resultants
--- (Expr,Expr) into FlatCurry function declarations and for
--- integrating these function declarations into a FlatCurry program
---
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module Prog2Flat(resultant2fundecl,fundecl2prog,fundecl2progA) where

import FlatCurry

-- transform a resultant into a function declaration:
resultant2fundecl :: (Expr,Expr) -> FuncDecl
resultant2fundecl (Comb FuncCall f vars, rhs) =
    Func f arity Public (genPolyType arity) (Rule (map (\(Var i)->i) vars) rhs)
    -- here we don't compute the real type of the function (TO DO!)
  where
    arity = length vars
    -- generate polymorphic type of the form (t1 -> t2 -> ... -> tn)
    -- where all ti's are different type variables:
    genPolyType n = if n==0 then TVar 0
                            else FuncType (TVar n) (genPolyType (n-1))


--creates a program from a set of function declarations:
fundecl2prog :: Prog -> [FuncDecl] -> Prog
fundecl2prog (Prog modname imports types oldfuns ops) newfuns =
  Prog modname imports types newfuns ops

--similar, but the functions of the original program are preserved:
--(observe that the first function is discarded because it also 
--appears with the same name in the specialization)
fundecl2progA :: Prog -> [FuncDecl] -> Prog
fundecl2progA (Prog modname imports types (_:oldfuns) ops) newfuns =
  Prog modname imports types (oldfuns++newfuns) ops
