------------------------------------------------------------------------------
--- This module implements the RLNT calculus 
--- (i.e., needed narrowing for flat programs)
--- 
--- @author JGuadalupe Ramos, Josep Silva, German Vidal
--- @version June 2005
------------------------------------------------------------------------------

module RLNT(unfold,rlnt,propBinding) where

import FlatCurry
import Unsafe

import General(hasGEN,generalize,delGEN2)

import Utils(substitute,substituteAll,index2vars,existsRen,
             getFuncDecls,maxList,maxVarIndex,isOpRooted,isCons,
             isConsRooted,maxList,isFlatVar)

import OffPretty(ppExpr,ppFormatList)

unfold :: Prog -> [Expr] -> Int -> Expr -> [Expr]

unfold prog memo fresh e 
  | isCons e       = [e]
  | isFlatVar e    = [e]
  | otherwise      = unfold_ prog memo fresh e --unfoldFun prog memo fresh e
  --if (isOpRooted e) 
    -- then unfold_ prog memo fresh e --unfoldFun prog memo fresh e
     --else [e]
          --unfoldC prog memo fresh e

unfoldC prog memo fresh e =
  if isCons e then [e]
              else case e of
                      Comb ConsCall c es -> concatMap (unfold prog memo fresh) es
                      _ -> unfold_ prog memo fresh e

unfoldFun prog memo fresh e =
  if b then [e]
       else unfold_ prog memo fresh e
  where (b,ne) = existsRen memo e

unfold_ prog memo fresh e =
  if ((isOpRooted e) && (hasGEN e)) then [e] --concatMap (unfold prog memo (fresh+1)) exps
                                    else unfold__ prog memo fresh e
--  where
--    exps = (generalize (fresh + (maxVarIndex e) + 1) e)

unfold__ prog memo fresh e =
  if e==e' then [e]
           else unfold prog memo' (fresh+1) e' 
  where (e',memo') = rlnt prog memo fresh e
              --(trace ("\n\nunfold__: " ++ (ppExpr 0 e) ++ "\nmemo: " ++ (ppFormatList (ppExpr 0) memo) ++ "\n") e)

rlnt :: Prog -> [Expr] -> Int -> Expr -> (Expr,[Expr])

-- Function call:
rlnt prog memo fresh (Comb FuncCall f es) =
  if hasGEN (Comb FuncCall f es)
     then (Comb FuncCall f es, memo)
     else rlntFC prog memo fresh (Comb FuncCall f es)

rlntFC prog memo fresh (Comb FuncCall f es) =
  if b || f==("Prelude","failed")
       then (Comb FuncCall f es, memo)
       else (matchRHS prog fresh f es, (Comb FuncCall f es) : memo)
  where (b,ne) = existsRen memo (Comb FuncCall f es)

--tt (fname,mod) = trace ("\n\nFuncCall: (" ++ fname ++ "," ++ mod ++ ")") (fname,mod)

--rlnt prog memo fresh (Comb FuncCall (mod,fname) es) = 
--  if fname=="GEN" then (Comb FuncCall (mod,fname) es, memo)
--                  else rlntF prog memo fresh (Comb FuncCall (mod,fname) es)

--rlntF prog memo fresh (Comb FuncCall f es) =
--  if b then (Comb FuncCall f es, memo)
--       else (matchRHS prog fresh f es, (Comb FuncCall f es) : memo)
--  where (b,ne) = existsRen memo (Comb FuncCall f es)

-- Var
rlnt prog memo fresh (Var x) = (Var x, memo)

-- Lits
rlnt prog memo fresh (Lit x) = (Lit x, memo)

-- Cons call:
rlnt prog memo fresh (Comb ConsCall f es) = 
  if isCons (Comb ConsCall f es) then (Comb ConsCall f es, memo)
                                 else (Comb ConsCall f nes, nmemo)
  where (nes,nmemo) = rlntSeq prog memo fresh es

--if hasGEN (Comb ConsCall f es) then (Comb ConsCall f es, memo)
--                               else (Comb ConsCall f nes, nmemo)

-- Case rules:

rlnt prog memo fresh (Case ctype e ces) =
  case e of
    Var n -> rlntCV prog memo fresh (Case ctype (Var n) (propBinding n ces))     --case guess
    Lit l -> (matchBranchLit ces l, memo)                       --case guess with a literal argument
    Comb ConsCall fname es -> (matchBranch ces fname es, memo)  --case select
    Comb FuncCall fname es -> rlnt_ prog memo fresh (Case ctype e ces)
    Case ctype2 (Var n) ces2 -> (Case ctype2 (Var n) (map (gen_subcase ctype ces) ces2), memo)  --case-of-case
    Case ctype2 ce2 ces2 -> rlnt__ prog memo fresh (Case ctype (Case ctype2 ce2 ces2) ces)     --case-of-case-of-nonvar

rlnt_ prog memo fresh (Case ctype e ces) =  (Case ctype e' ces, memo')
  where (e',memo') = rlnt prog memo fresh e

rlnt__ prog memo fresh (Case ctype e ces) = (Case ctype e' ces, memo')
  where (e',memo') = rlnt prog memo fresh e

rlntCV prog memo fresh (Case ctype var ces) = (Case ctype var nces, nmemo)
  where (nces,nmemo) = rlntSeqBranch prog memo fresh ces
  

--auxiliary function for case-of-case
gen_subcase :: CaseType -> [BranchExpr] -> BranchExpr -> BranchExpr
gen_subcase ctypej cesj (Branch pat ek) = Branch pat (Case ctypej ek cesj)


-- match a branch (for case select):
matchBranch :: [BranchExpr] -> QName -> [Expr] -> Expr
matchBranch [] _ _ = (Comb FuncCall ("Prelude","failed") [])
matchBranch (Branch (Pattern p vars) e : ces) c es =
  if p==c then substitute vars es e else matchBranch ces c es

-- match a branch with Literals:
matchBranchLit :: [BranchExpr] -> Literal -> Expr
matchBranchLit [] _ = (Comb FuncCall ("Prelude","failed") [])
matchBranchLit (Branch (LPattern p) e : ces) c =
  if p==c then e else matchBranchLit ces c

--propagate matchings to the corresponding branches (for case guess):
propBinding :: Int -> [BranchExpr] -> [BranchExpr]

propBinding _ [] = []
propBinding v (Branch (Pattern c vs) e : es) = 
   (Branch (Pattern c vs) 
           (substitute [v] [Comb ConsCall c (index2vars vs)] e)) : propBinding v es
propBinding v (Branch (LPattern (Intc p)) e : es) = 
   (Branch (LPattern (Intc p))
           (substitute [v] [Lit (Intc p)] e)) : propBinding v es
propBinding v (Branch (LPattern (Floatc p)) e : es) = 
   (Branch (LPattern (Floatc p))
           (substitute [v] [Lit (Floatc p)] e)) : propBinding v es
propBinding v (Branch (LPattern (Charc p)) e : es) = 
   (Branch (LPattern (Charc p))
           (substitute [v] [Lit (Charc p)] e)) : propBinding v es

-- match a right-hand side of a given function (for fun eval):
matchRHS :: Prog -> Int -> QName -> [Expr] -> Expr
matchRHS prog fresh f es = getMatchedRHS (getFuncDecls prog) fresh f es

getMatchedRHS :: [FuncDecl] -> Int -> QName -> [Expr] -> Expr
getMatchedRHS (Func fname _ _ _ funrule : fds) fresh name es =
   if fname==name then getMatchedRHS_aux funrule fresh es
                  else getMatchedRHS fds fresh name es

getMatchedRHS_aux :: Rule -> Int -> [Expr] -> Expr
getMatchedRHS_aux (Rule vars rhs) fresh es =
  let 
      maxindex = (max (max (maxList vars) (maxVarIndex rhs)) (maxList (map maxVarIndex es))) + (fresh * 100)
  in
    substituteAll vars es maxindex rhs


--(trace ("\nvars: " ++ (ppFormatList show vars) ++ "\n") vars) (trace ("\nes: " ++ (ppFormatList (ppExpr 0) es) ++ "\n") es) (trace ("\nmaxindex: " ++ show maxindex ++ "\n") maxindex) (trace ("\nrhs: " ++ (ppExpr 0 rhs) ++ "\n") rhs)

--sequential application of rlnt (map cannot be used because of the need
--to update the memo table)

rlntSeq prog memo fresh [] = ([],memo)
rlntSeq prog memo fresh (e:es) = (ne:nes,nnmemo)
  where (ne,nmemo) = rlnt prog memo fresh e
                               --(trace ("\n\rlntSeq (e): " ++ (ppExpr 0 e) ++ "\n") e)
        (nes,nnmemo) = rlntSeq prog nmemo fresh es

rlntSeqBranch prog memo fresh [] = ([],memo)
rlntSeqBranch prog memo fresh (Branch p e:es) = (Branch p ne : nes, nnmemo)
  where (ne,nmemo) = rlnt prog memo fresh e
                                      --(trace ("\nrlntSeqBranch (e): " ++ (ppExpr 0 e) ++ "\n") e)
        (nes,nnmemo) = rlntSeqBranch prog nmemo fresh es