module afpgraph(getGraph, union) where

import fptypes         -- It includes graph definitions
import AbstractCurry   -- It defines AbstractCurry structures


------------------------------------------------------------------------------------------
--  jg ramos, g arroyo,j silva, g vidal  23 agosto 2006
--  getGraph, collect a list of functions and function calls  [("f",["g","h","i"],_)]
--  from an AbstractCurry program then
--  analize the resulting list and mark with C = cyclic and NC= non cyclic each function
--  It returns a kind of type Graph = [("f",["g","h","i"],C), ... ]
------------------------------------------------------------------------------------------
getGraph:: CurryProg -> Graph
getGraph prog =  let gr1  = scan prog emptygR  
                 in  markG gr1

------------------------------------------------------------------------------------------
-- scan receives an AbstractCurry Program and returns a graph like: [("f",["g","h","i"],_)]
-- but still do not know if the function is cyclic or non cyclic
------------------------------------------------------------------------------------------


scan :: CurryProg -> Graph -> Graph
scan (CurryProg _ _ _ funs _ ) gR = functionIn funs gR

           
--functionIn ::[CFuncDecl] -> Graph -> Graph
--functionIn  []    gR = gR
--functionIn (f:fs) gR = map(checkAF gR) fs
          
functionIn ::[CFuncDecl] -> Graph -> Graph
functionIn  []    gR = gR
functionIn (f:fs) gR = checkAF f (functionIn fs gR)


-- add a new node if the visited function if it is not in the graph

checkAF :: CFuncDecl -> Graph -> Graph
checkAF (CFunc (_,fname) _ _ _ (CExternal _) ) gR = gR 
checkAF (CFunc (_,fname) _ _ _ (CRules cev rules) ) gR = 
      checkR rhs fname ( if (existNode fname gR) then gR else (addNode fname gR) )


-------- ------>   inF  --> Name of function that contains the rules

checkR:: Expr -> String  -> Graph -> Graph
checkR e inF gR =
 case e of
  Free _ exp             ->        checkR      exp inF gR
  Comb ConsCall _     es ->        checkSubCons es inF gR
  Comb (FuncPartCall _) _ es ->        checkSubCons es inF gR
  Comb FuncCall (_ ,fcall)  es  -> checkSubCons es inF (regCall fcall inF gR ) 
  Case _ c bs                   -> let gg = (checkR c inF gR) in checkBranches bs inF gg
  Or ex ey                      -> checkSubCons [ex,ey] inF gR
  _                             -> gR


checkSubCons::[Expr]-> String -> Graph -> Graph
checkSubCons []     _   gR = gR
checkSubCons (e:es) inF gR = checkR e inF gg
              where gg = checkSubCons es inF gR

checkBranches :: [BranchExpr] -> String -> Graph -> Graph
checkBranches [] _  gR = gR
checkBranches ( (Branch _ b):bs ) inF gR = checkR b inF gg
          where gg = checkBranches bs inF gR



-- Next THREE functions are OPTIMIZABLE, they add only calls
----------------------------------------------------------------------------------------
--1  register a fcall for the function inF
---------------------------------------------------------------------------------------- 
regCall fcall inF gR = if (existCall fcall inF gR) then gR
                       else (addCall fcall inF gR)


-----------------------------------------------------------------------------------------
-- 2 if there is a function call fcall registered for the previously added 
-- function inF (since checkF ), this checks in all graph
-----------------------------------------------------------------------------------------
existCall fcall inF ((f,lcalls,_):gs) = 
    if (f==inF) then (viewCalls fcall lcalls)  else (existCall fcall inF gs)
    where
       viewCalls _  [] = False
       viewCalls fc (c:cs)= if (c==fc) then True else (viewCalls fc cs)


-----------------------------------------------------------------------------------------
--           3 add a new fcall for function inF, check all graph
-----------------------------------------------------------------------------------------
addCall fcall inF ((f,lcalls,ch):gs) =
    if (f==inF) then ((f,newLCalls,ch):gs) else ((f,lcalls,ch):(addCall fcall inF gs))
    where
       newLCalls = concat [lcalls, [fcall] ]


-----------------------------------------------------------------------------------------
-- Checks if a function (f,,) is in the graph
-----------------------------------------------------------------------------------------
existNode _  []     = False
existNode fname ((fn,_,_):gs) = if (fn==fname) then True 
                                else existNode fname gs 

----------------------------------------------------------------------------------------
-- Add a new node (f,,) on the graph
----------------------------------------------------------------------------------------
addNode:: String->Graph->Graph
addNode fname gR = concat [gR, [(fname,[],NC)] ]




------------------------------------------------------------------------------------------
-----********************************************************************************-----
-- markG
-- It proceeds enabling (C)cyclic  or (NC) not cyclic functions
------------------------------------------------------------------------------------------

markG graph = let 
                  g = markGraph1 graph 
              in markGraph2 g


-------------------- This section discovers really cyclic functions------------------

markGraph1 graph = map (analizeF graph) graph


--------------------------------------------------------------------------------------
-- This function asks the function call universe, if f is there then is a cyclic funct
--------------------------------------------------------------------------------------
analizeF graph (f,cs,_) = if (cyclic f cs) then (f,cs,C) else (f,cs,NC)
       where
         cyclic fn calls = elem fn (universeF calls graph)


--------------------------------------------------------------------------------------
-- Universe is the list that has the functions towards which the function g conduces
-- universeF returns the universe of F, making multiple passes
--------------------------------------------------------------------------------------
universeF calls g = 
              if (length calls == length newlist) then calls
              else (universeF newlist g )
            where newlist = getUnivIns calls g



-------------------------------------------------------------------------------------
-- getUnivIns, returns the call universe of a function, without repeated f's,
-- we suposse calls do not has repated f's because when we construct the 
-- graph (scan) we forbid that
-------------------------------------------------------------------------------------
getUnivIns calls g = foldl union calls (getCalls calls g)

--------------------------------------------------------------------------------------
-- getCalls, in one pass gets the function names that the original list conduces
-- e.g. ("f",["g","h"],_), if (g) conduce to (i) and (h) to (j) then getUnivIns
-- gives ["g","i","h","j"] to "f"
--------------------------------------------------------------------------------------

getCalls []     _ = []
getCalls (x:xs) g = getCallsN x g : getCalls xs g


--------------------------------------------------------------------------------------  
-- getCallsNode returns the function calls from f

getCallsN f graph  = ln
     where
        (_,ln,_) = getNode f graph


--------------------------------------------------------------------------------------
-- getNode returns a graph node that contains name indicated
-- if the called function is not in graph, that is a built in function
--------------------------------------------------------------------------------------
getNode  _   []          = ("",[],NC)
getNode  f ((f1,l,c):xs) = if (f==f1) then (f1,l,c) else getNode f xs

-------------------------------------------------------------------------------------
-- Given a function list, and their corresponding appearing function calls,it returns
-- a list of functions without repeated function names
-------------------------------------------------------------------------------------


union list1 list2 = runion (concat [list1,list2]) []

runion   []   l2  = l2
runion (x:xs) l2 = if (elem x l2) then runion xs l2
                  else runion xs  (x:l2)


-----------------------------------------------------------------------------------         
-- if a cyclic function "f" conduces to a non cyclic "g", then g is cyclic
-----------------------------------------------------------------------------------
markGraph2bis graph = map (analizeF2 collection) graph
          where
             collection = foldr union [] (collect graph)


markGraph2 graph = if (countCyclic graph == countCyclic newGraph) 
                       then graph
                   else (markGraph2 newGraph )
                         where 
                            newGraph = markGraph2bis graph

countCyclic [] = 0
countCyclic ((_,_,c):gs) = if (c==C) then (1 + countCyclic gs)
                           else countCyclic gs

-----------------------------------------------------------------------------------
-- collect returns a list of called functions from cyclic functions
-----------------------------------------------------------------------------------
collect  [] = []
collect ((_,l,c):gs) = if (c == C) then l:collect gs
                       else collect gs


-----------------------------------------------------------------------------------
-- if it is clear that f is called from cyclic then is cyclic, else we leave it with
-- the original creterion
----------------------------------------------------------------------------------

analizeF2 coll (f,cs,c) = if (elem f coll) then (f,cs,C)
                           else (f,cs,c)

