-----------------------------------------------------------------------------------------
-- A simple binding-time analysis for AbstractCurry programs V. 0.1
-- 
-- Gustavo Arroyo, J.Guadalupe Ramos, Josep Silva, Germán Vidal
--  Valencia, Sept., 2006
-----------------------------------------------------------------------------------------
module btaAbst where

import AbstractCurry
import Unsafe
import List

data BindingTime = S | D
type BTEnv       = [BindingTime]
type Div         = [BTEnv]  
data Variable    = VarT Int
type Fname       = String
type Arity       = Int
type CBTEnv      = [(Variable,BindingTime)]
type BTD         = [(Fname,Arity,BTEnv)]    --- BindingTime Division     
                                                  

        -- call with:   bta "program" tau     
        -- where tau is a list of BT values for the first function  of the program, 
        -- 
        -- example:>  bta "power" [D,S]
        -- 


callbta :: String -> IO()
callbta fichero = bta fichero initDiv
        where initDiv = []

bta :: String -> BTEnv -> IO()
bta  p tau0 = 
  do printVersion
     prog <- readCurry (p) 
     let  (gas,div) = explore prog tau0 
     let out        = zipWith makeFinD gas div
     putStrLn $  "... Division has been calculated: \n" ++ ("\n"++(show out))


             --regresa una tupla que contiene una lista de tuplas (f,a) y una division
explore :: CurryProg -> BTEnv -> ([(String,Int)],Div)                      
explore (CurryProg _ _ _ fs _) tau0 = ((g_a), (calcDiv tau0 g_a div0 funs))
    where g_a  =  getArities fs
          div0 =  tau0: (tail (map statictau (map snd g_a)))
          funs =  getUsrFun fs

makeFinD  (g,a) taug = (g,a,taug)

                                               -- it returns (function,arity)
getArities []  = []
getArities ((CFunc (_,f) a _ _ rules):fs) = getArities_ rules
                 where
                    getArities_ (CRules  _  _ ) = if a>0 then (f,a):getArities fs else getArities fs
                    getArities_ (CExternal _ )  = getArities fs

                                              --it returns a CFuncDecl iff
getUsrFun [] = []  
getUsrFun ((CFunc qn a v t rules):fs) = getUsrFun_ rules
               where
                 getUsrFun_ (CRules _  _ ) = if a>0 then (CFunc qn a v t rules):getUsrFun fs else getUsrFun fs
                 getUsrFun_ (CExternal _ ) = getUsrFun fs

--- Calcula la division del programa dados: BTenv inicial, lista de g_as, division inicial y una [CFuncDecl]
        ---  g_as [(function,arity)] [CFuncDecl]
calcDiv tau0 g_as div0               funcs      = if (eqDiv div0 div1) then  div1           
                                                    else  calcDiv tau0 g_as div1 funcs 
                    where (d:ds) = compDiv g_as div0 funcs
                          div1 = (lubTau d tau0):ds


eqDiv xs ys = not (elem False (zipWith eqTau xs ys))


compDiv    []   _   _  = []                                                   
compDiv ((g,a):gns) div fs = (foldr1 lubTau srchGinD) : compDiv gns div fs
                     where 
                       srchGinD = divGinFs g a div fs

                   --- searches a function call in all functions with all taus, returns a tau of arity a
divGinFs  _    _   []       []  = [] 
divGinFs gname a (tau:taus) ((CFunc _ _ _ _ crules) : fs) =
                  (divGinF gname a tau crules):(divGinFs gname a taus fs)



divGinF gname a tau (CRules _ lcrule) = foldr lubTau (statictau a) (map (divGinF_ gname a tau) lcrule)

--- no acepta where anidado (solo where, funcion o local free variable) 


divGinF_ g a tau (CRule lcpat [(_,cexp)] lCDcl) =
      case lCDcl of
                 [] -> Bv ctau g a cexp
                 _  -> foldr lubTau (Bv ctau g a cexp) (map (bvletwhere tau g a) lCDcl)
        where ctau = createctau tau lcpat

-- Para el caso de [CLocalDecl] no vacio tomamos el mismo ctau de la función donde llama
-- a bvletwhere 
bvletwhere tau gname a clocDcl =
    case clocDcl of 
          (CLocalVar _ )                      -> statictau a
          (CLocalFunc (CFunc _ _ _ _ crules))-> divGinF gname a tau crules
--          (CLocalPat cpat cexpr _)  -> let ctau = createctau tau cpat  in Bv ctau gname a cexpr
--          (CLocalPat cpat cexpr _)            -> Bv ctau gname a cexpr         ---where


toCtau vidx btval = (VarT vidx, btval)


-------------------------------------------------------------------------------------------------
--  Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be 
-------------------------------------------------------------------------------------------------
------------------------------------------------------
-- Input:  expr, the expression to be analized, 
--         ctau, curry tau
-- Output: a btvalue
------------------------------------------------------

Be :: CBTEnv -> CExpr  ->   BindingTime 
Be ctau expr  =
  case expr of
     (CVar     v)        -> tauOf ctau (CVar v)  
     (CLit    _ )        -> S
     (CSymbol _ )        -> S
     (CApply le re)      -> lub (Be ctau le) (Be ctau re)     
     (CCase (CVar x) bs) -> let btv = tauOf ctau (CVar x)              
                                 in  foldr lub btv (map (processBranch ctau (CVar x)) bs)


processBranch ctau  varCase (CBranch (CPVar cvariname) expr) =   Be ctau1 expr
               where ctau1 =  ctau ++ [(toVar (tauOf ctau varCase) cvariname)] 
processBranch ctau  _       (CBranch (CPLit _ ) expr)        =   Be ctau  expr
processBranch ctau  varCase (CBranch (CPComb _ lcpat) expr)  =   Be ctau2 expr
               where ctau2 =  ctau ++ (map (toVarlpat (tauOf ctau varCase)) (flatlcpat lcpat))



flatlcpat lcpat =
   case lcpat of
      ((CPVar cvin):[])          -> (CPVar cvin):[]
      ((CPVar cvin):rlcpat)      -> (CPVar cvin):(flatlcpat rlcpat)
      ((CPComb _ []):[])         -> []
      ((CPComb _ ilcpat):[])     -> (flatlcpat ilcpat)
      ((CPComb _ []):rlcpat)     -> (flatlcpat rlcpat)
      ((CPComb _ ilcpat):rlcpat) -> (flatlcpat ilcpat)++(flatlcpat rlcpat)


-- en este caso lcpat debe contener una lista de (CPVar CVarIName)
--------------------------------------------------------
-- Input: btval, a binding time value
--        var,   a case internal variable
-- Return a pair: (Var n, btval)
--------------------------------------------------------
toVar:: BindingTime -> CVarIName -> (Variable,BindingTime)
toVar btval (index,_) = (VarT index, btval)

--toVarlpat:: BindingTime -> CPattern -> (Variable,BindingTime)
toVarlpat btval (CPVar (index, _)) = (VarT index, btval)
--toVarlpat btval (CPComb _ []) = []
-------------------------------------------------------------------------------------------------
--  Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv 
-------------------------------------------------------------------------------------------------

    --curryTau,  searchedG, aritiyG,   Expression
Bv ::  CBTEnv  ->  Fname  -> Arity  ->  CExpr  ->  BTEnv 
Bv ctau g a expr =
  case expr of
     (CVar   _ )            -> statictau a
     (CLit   _ )            -> statictau a
     (CSymbol _ )           -> statictau a
     (CApply _ _)           -> evalapply ctau g a (plBCE (reverse (Bvint ctau g expr)))
     (CDoExpr lstatmt)      -> foldr lubTau (statictau a) (map (procstatmt ctau g a) lstatmt) 
     (CLetDecl _ cexp)      -> Bv ctau g a cexp           --caso de un let dentro de una regla
     (CCase (CVar x)   bs ) -> let xtau = statictau a              
                                    in foldr lubTau xtau (map (processBranchTau ctau g a (CVar x)) bs)

Bvint ctau g expr =
      case expr of
           (CVar _)        -> [(False, expr)]
           (CLit _)        -> [(False, expr)]
           (CSymbol (_,f)) -> let t = [(False, expr)] in if f==g then [(True,expr)] else t
           (CApply le re)  -> (False, re): (Bvint ctau g le)


plBCE ((b,_):rlst) = if b then (map (putT) rlst )
                          else  rlst

putT (_,e) = (True,e)

evalapply _    _ _ []            = []
evalapply ctau g a ((b,e):rlBCE) = if b then (Be ctau e):evalapply ctau g a rlBCE
                                     else foldr lubTau (statictau a) (map (eval_eBCE ctau g a) ((b,e):rlBCE))

eval_eBCE ctau g a (_,e) = Bv ctau g a e 

procstatmt ctau g a (CSExpr cexp) = Bv ctau g a cexp
procstatmt ctau g a (CSPat _ cexp) = Bv ctau g a cexp
--procstatmt ctau g a (CSLet lcdec) = foldr lubTau (statictau a) (map (bvletwhere ctau g a) lcdec)



processBranchTau ctau g a varCase (CBranch (CPVar cvariname) expr) =   Bv ctau1 g a expr
                       where ctau1 = ctau ++ [(toVar (tauOf ctau varCase) cvariname)]
processBranchTau ctau g a _       (CBranch (CPLit _ ) expr)        =   Bv ctau g a expr
processBranchTau ctau g a varCase (CBranch (CPComb _ lcpat) expr)  =   Bv ctau2 g a expr
                       where ctau2 = ctau ++ (map (toVarlpat (tauOf ctau varCase)) lcpat)

-- en este caso lcpat debe contener una lista de (CPVar CVarIName)

------------------------------------------------------------------------------
--                  TOOLS
------------------------------------------------------------------------------

------------------------------------------------------------
-- Input: 
-- ctau :the expression to be analized, 
-- Expr : a 
------------------------------------------------------------

tauOf :: CBTEnv -> CExpr -> BindingTime 
tauOf   []               _      = D
tauOf (((VarT v),bt):xs) (CVar (x,cvar)) = if x==v then bt
                                   else tauOf xs (CVar (x,cvar)) 

btOflct :: CBTEnv -> Int -> BindingTime 
btOflct   []               _  = D
btOflct (((VarT v),bt):xs) x  = if x==v then bt
                                   else btOflct xs x 

lubTau xs ys = zipWith (lub) xs ys                                          

lub :: BindingTime -> BindingTime -> BindingTime
lub  x y = arit x y

arit :: BindingTime -> BindingTime -> BindingTime
arit  S  S = S
arit  S  D = D
arit  D  S = D
arit  D  D = D

statictau a =  map (\_ -> S) [1..a]

eqTau xs yx = not (elem False (zipWith eqbt xs yx))

eqbt :: BindingTime -> BindingTime -> Bool
eqbt  S  S = True
eqbt  S  D = False
eqbt  D  S = False
eqbt  D  D = True

----------------------------------------------------------------------
printVersion =
   do putStrLn " "
      putStrLn "---- calculating BTA ----"       

---------------------------------------------------------------------
flatcexpw :: CExpr -> [CExpr]
flatcexpw cexp      = nub (flatcexp cexp)


flatcexp :: CExpr -> [CExpr]
flatcexp cexp =
   case cexp of
      (CVar cvin)          -> (CVar cvin):[]
      (CLit _)             -> []
      (CSymbol _ )         -> []
      (CApply le re)       -> (flatcexp le)++(flatcexp re)
      (CLetDecl _ ce)      -> (flatcexp ce)
      (CDoExpr lstat)      -> (flatlstat  lstat)
      (CListComp ce lstat) -> (flatcexp ce)++(flatlstat lstat)
      (CCase ce lbrex)     -> (flatcexp ce)++(flatlbrex lbrex)

flatlstat []           = []
flatlstat (stat:rstat) =  (flatstat stat)++(flatlstat rstat)

flatstat (CSExpr cexp)  = flatcexp cexp
flatstat (CSPat _ cexp) = flatcexp cexp
flatstat (CSLet lcdec)  = foldr (++) [] (map (flatletwhere ) lcdec)


flatletwhere clocDcl =
    case clocDcl of 
        (CLocalPat _ cexpr _)  -> flatcexp cexpr         ---where
        (CLocalVar _ )         -> []
        (CLocalFunc _)         -> []

flatlbrex []      = []
flatlbrex (brex:rbrex) = (flatbrex brex)++(flatlbrex rbrex)

flatbrex (CBranch _  cexp)   = flatcexp cexp

createbtEindx :: [CPattern]-> BTEnv ->  [(CPattern,BindingTime)]
createbtEindx lcpat btE = zip lcpat btE

createctau tau lcpat = concat (createlVarTBT (createbtEindx lcpat tau))

createlVarTBT  []                 = []
createlVarTBT  (cpatbtE:rcpatbtE) = (createVarTBT cpatbtE):(createlVarTBT rcpatbtE)

createVarTBT ((CPVar (i,_)),bind)     = ((VarT i), bind):[]
createVarTBT ((CPComb _ lcpat),bind) = foldr (++) [] (map (creaVarTBT bind) lcpat)


creaVarTBT  bind (CPVar (i,_))    =  ((VarT i), bind):[]
creaVarTBT  bind (CPComb _ lcpat) =  foldr (++) [] (map (creaVarTBT bind) lcpat)

--creactaulD tau cpat = concat (createlVarTBT (createbtEindx lcpat tau))
