-----------------------------------------------------------------------------------------
-- A simple binding-time analysis for flatCurry programs V. 0.1
-- 
-- Gustavo Arroyo, J.Guadalupe Ramos, Josep Silva, Germán Vidal
--  Valencia, April 28th, 2006
-----------------------------------------------------------------------------------------
                                                        --gad 30ago06
module btaFlat where

import FlatCurry
import Unsafe

data BindingTime = S | D
type BTEnv       = [BindingTime]
type Div         = [BTEnv]  
data Variable    = VarT Int
type Fname       = String
type Arity       = Int
type CBTEnv      = [(Variable,BindingTime)]       
type BT          = [(Fname,Arity,BTEnv)]                                                  

        -- call with:   bta "program" tau     
        -- where tau is a list of BT values for the first function  of the program, 
        -- 
        -- example:>  bta "power" [D,S]
        -- 


callbta :: String -> IO()
callbta fichero = bta fichero initDiv
        where initDiv = []



bta :: String -> BTEnv -> IO()
bta  p tau0 = 
  do printVersion
     prog <- readFlatCurry p 
     let  (gas,div) = explore prog tau0 
     let out        = zipWith makeFinD gas div
     putStrLn $  "... Division has been calculated: \n" ++ ("\n"++(show out))



explore :: Prog -> BTEnv -> ([(String,Int)],Div)                      
explore (Prog _ _ _ fs _) tau0 = ((g_a), (calcDiv tau0 g_a div0 funs))
    where g_a  =  getArities fs
          div0 =  tau0: (tail (map statictau (map snd g_a)))
          funs =  getUsrFun fs

makeFinD  (g,a) taug = (g,a,taug)

getArities [] = []                                                 --it returns (function,arity)
getArities ((Func (_,f) a _ _ rules):fs) = getArities_ rules
                 where
                    getArities_ (Rule  _  _ ) =  if a>0 then (f,a):getArities fs else getArities fs
                    getArities_ (External _ ) =  getArities fs

getUsrFun [] = []                                                          
getUsrFun ((Func qn a v t rules):fs) = getUsrFun_ rules
               where
                 getUsrFun_ (Rule _   _ ) = if a>0 then (Func qn a v t rules):getUsrFun fs else getUsrFun fs
                 getUsrFun_ (External _ ) = getUsrFun fs


calcDiv tau0 g_as div0 funcs = if (eqDiv div0 div1) then  div1           
                                 else  calcDiv tau0 g_as div1 funcs 
                    where (d:ds) = compDiv g_as div0 funcs
                          div1  = (lubTau d tau0):ds   


eqDiv xs ys = not (elem False (zipWith eqTau xs ys))


compDiv    []   _   _  = []                                                   
compDiv ((g,a):gns) div fs = (foldr1 lubTau srchGinD) : compDiv gns div fs
                     where 
                       srchGinD =  divGinFs g a div fs

                              --- searches a function call in all functions with all taus, returns a tau of arity a
divGinFs  _    _   []       []  = [] 
divGinFs gname a (tau:taus) ( (Func _ _ _ _ (Rule vidxs expr)) : fs ) = 
                    (Bv ctau gname a expr):(divGinFs gname a taus fs)        
              where ctau = zipWith toCtau vidxs tau

--Bv ctau g a expr

toCtau vidx btval = (VarT vidx, btval)



-------------------------------------------------------------------------------------------------
--  Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be Be 
-------------------------------------------------------------------------------------------------
------------------------------------------------------
-- Input:  expr, the expression to be analized, 
--         ctau, curry tau
-- Output: a btvalue
------------------------------------------------------

Be :: CBTEnv -> Expr  ->   BindingTime 
Be ctau expr  =
  case expr of
     (Var      v)           -> tauOf ctau (Var v)  
     (Lit     _ )           -> S
     (Comb ConsCall _ args) -> foldr lub S (map (Be ctau) args)         
     (Comb FuncCall _ args) -> foldr lub S (map (Be ctau) args) 
     (Case _ (Var x)   bs ) -> let btv = tauOf ctau (Var x)               -- case p/fs Inductively Seq
                                   in  foldr lub btv (map (processBranch ctau (Var x)) bs)
     (Or e1 e2)             -> lub (Be ctau e1) (Be ctau e2)
 
-- faltan branch literales

processBranch ctau varCase (Branch (Pattern _ varIdx) expr) =   Be ctau2 expr
   where  ctau2 =   ctau ++ (map (toVar (tauOf ctau varCase)) varIdx)

--------------------------------------------------------
-- Input: btval, a binding time value
--        var,   a case internal variable
-- Return a pair: (Var n, btval)
--------------------------------------------------------
toVar:: BindingTime -> Int -> (Variable,BindingTime)
toVar btval var = (VarT var, btval)


-------------------------------------------------------------------------------------------------
--  Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv Bv 
-------------------------------------------------------------------------------------------------


    --curryTau,  searchedG, aritiyG,   Expression
Bv ::  CBTEnv  ->  Fname  -> Arity  ->  Expr  ->  BTEnv 
Bv ctau g a expr =
  case expr of
     (Var   _ )                 -> statictau a
     (Lit   _)                  -> statictau a
     (Comb ConsCall _ args    ) -> foldr lubTau (statictau a) (map (Bv ctau g a) args)
     (Comb FuncCall (_,f) args) -> let t = foldr lubTau (statictau a) (map (Bv ctau g a) args)
                                       in if (f==g) then lubTau t (map (Be ctau) args) 
                                                else t
     (Case _ (Var x)   bs ) -> let xtau = statictau a               -- case p/fs Inductively Seq
                                   in foldr lubTau xtau (map (processBranchTau ctau g a (Var x)) bs)

processBranchTau ctau g a varCase (Branch (Pattern _ varIdx) expr) =   Bv ctau2 g a expr
   where  ctau2 =   ctau ++ (map (toVar (tauOf ctau varCase)) varIdx)



------------------------------------------------------------------------------
--                  TOOLS
------------------------------------------------------------------------------

------------------------------------------------------------
-- Input: 
-- ctau :the expression to be analized, 
-- Expr : a 
------------------------------------------------------------

tauOf :: CBTEnv -> Expr -> BindingTime 
tauOf   []               _      = D
tauOf (((VarT v),bt):xs) (Var x) = if x==v then bt
                                   else tauOf xs (Var x) 

lubTau xs ys = zipWith (lub) xs ys                                          

lub :: BindingTime -> BindingTime -> BindingTime
lub  x y = arit x y

arit :: BindingTime -> BindingTime -> BindingTime
arit  S  S = S
arit  S  D = D
arit  D  S = D
arit  D  D = D

statictau a =  map (\_ -> S) [1..a]

eqTau xs yx = not (elem False (zipWith eqbt xs yx))

eqbt :: BindingTime -> BindingTime -> Bool
eqbt  S  S = True
eqbt  S  D = False
eqbt  D  S = False
eqbt  D  D = True

----------------------------------------------------------------------
printVersion =
   do putStrLn " "
      putStrLn "---- calculating Jones' BTA ----"       
--      putStrLn "----Jones BTA for  FlatCurry----"
--      putStrLn "---------- oct-nov 05 ----------" 
--      putStrLn " "

---------------------------------------------------------------------
--tests                                                
--t1= be (Var 1) [(Vt 0,D),(Vt 1, S),(Vt 2,S)]

t1=  Be [(VarT 0,D),(VarT 1,S)] (Case Flex (Var 0) [(Branch (Pattern ("menorque","Z") []) (Comb ConsCall ("prelude","True") [])),(Branch (Pattern ("menorque","S") [2]) (Case Flex (Var 1) [(Branch (Pattern ("menorque","Z") []) (Comb ConsCall ("prelude","False") [])),(Branch (Pattern ("menorque","S") [3]) (Comb FuncCall  ("menorque","menorque") [(Var 2),(Var 3)]))]))])



t2=  Bv [(VarT 0,D),(VarT 1,S)] "menorque" 2 (Case Flex (Var 0) [(Branch (Pattern ("menorque","Z") []) (Comb ConsCall ("prelude","True") [])),(Branch (Pattern ("menorque","S") [2]) (Case Flex (Var 1) [(Branch (Pattern ("menorque","Z") []) (Comb ConsCall ("prelude","False") [])),(Branch (Pattern ("menorque","S") [3]) (Comb FuncCall  ("menorque","menorque") [(Var 2),(Var 3)]))]))])

