-- This module implements qs(l,r)

module fpqs (qsF) where

import fptypes
import FlatCurry

---------------------------------------------------------------------------------------
-- jg ramos, 13 dic 2004
--
-- qsF receives a function data structure in flat curry, further marks with the bullet 
-- function the dangerous function arguments (increasing or nested function calls)
--
-- g. arroyo, 11 ago 2006
-- se ha adaptado el anotador "no creciente" de acuerdo con el especializador RLNT,
-- donde este último toma en cuenta anotaciones con la definicion "GEN" en lugar de 
-- la definición "bullet"
---------------------------------------------------------------------------------------
qsF :: FuncDecl -> FuncDecl
qsF (Func  qn  ar t v (External st   ) )  = (Func qn ar t v (External st       ) )   
qsF (Func  qn  ar t v (Rule rule prhs) )  = (Func qn ar t v (Rule rule newprhs ) )  
  where
    newprhs = qs (toVD 0 rule)   0      rule  prhs
              -- rule->varDepth, depth, rule (positions of cases Var)


---------------------------------------------------------------------------------------
--  qs view, specially, function calls and verifies its arguments to be aware with
--  variable depth
---------------------------------------------------------------------------------------

      -- varlist -- depth -- rule -- case+body--newexpr 
qs::  [VarDepth] -> Int -> [VarIndex]-> Expr -> Expr
qs    vars  currd  rule e =
  case e of
    Case ct (Var n) bs  -> Case ct (Var n) (processCase bs n vars currd rule)
    Var x               -> Var x
    Comb ConsCall qn es -> Comb ConsCall qn (map (qs  vars currd rule) es)
    Comb FuncCall qn es -> Comb FuncCall qn (map (qs2 vars qn       ) es) 
    _                   -> e

qs2 :: [VarDepth] -> QName ->Expr -> Expr
qs2 vardepthP (prog,_) t =
     if ( (isCons t) && (and (map (compareVar t) vardepthP)) ) then t
     else Comb FuncCall (prog,"GEN") [(qs vardepthP 0 [] t)]
--     else Comb FuncCall (prog,"bullet") [(qs vardepthP 0 [] t)]      --gad11ago06


compareVar t (x,d) = if (d >= (depthV x t)) then True else False


----------------------------------------------------------------------------
-- process case, do not return code, only inits depth or detects non variables
----------------------------------------------------------------------------
processCase:: [BranchExpr]->VarIndex->[VarDepth]->Int->[VarIndex]->[BranchExpr]
processCase bs n vars currd rule = 
        let                        --si este case VAR se encuentra en vars, quitarlo
           vs = if (elem2 n vars) then (delVn n vars) else vars
           d  = if (elem  n rule) then 0 else currd   --iniciar cuenta en 0
        in  checkBrs bs vs d rule 


----------------------------------------------------------------
checkBrs :: [BranchExpr] -> [VarDepth] -> Int ->[VarIndex] -> [BranchExpr]
checkBrs [] _ _  _ = []
checkBrs ((Branch (Pattern qn varB) e) : bs) vars depth rule  =
 (Branch (Pattern qn varB) newE):checkBrs bs vars depth rule 
      where 
        d     = depth+1                   --inc depth and add new vars, if varB not null
        newvs = concat [ vars , (toVD d varB) ]
        newE  = if (null varB) then qs vars  depth rule e -- case  Z -> e
                else                qs newvs d     rule e --       Sx-> e       

------------------------------------------------------------------------
-- toVD d vs, pass from n,[2,4,5] to [(2,n),(4,n)(5,n)]
------------------------------------------------------------------------
toVD d vs = map (putDv d) vs
                   
putDv d x = (x,d)

------------------------------------------------------------------------
-- elem2 view if a var n is in [(nvar, depth)]
------------------------------------------------------------------------
elem2 :: Int -> [VarDepth] -> Bool
elem2 _   [ ]      = False
elem2 n ((v,_):vs) = if (v==n) then True else (elem2 n vs)

------------------------------------------------------------------------
-- del if the var n is in [(nvar, depth)]
------------------------------------------------------------------------
delVn _    []      = []
delVn n ((v,d):vs) = if (n==v) then vs else((v,d):(delVn n vs))




--*-*-*-*-*-*-*-*-*-*-*-*- DEPTH  MEASURING    CODE  -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*--
---------------------------------------------------------------------------------------
-- depthV, encounters the variable depth in a constructor expresion
---------------------------------------------------------------------------------------

depthV :: Int ->  Expr-> Int
depthV v e =                --it runs well still when x is not an elem of a constructor term
 case e of                  
  Var  n            -> if(n==v) then 0  else (-1)    
  Comb ConsCall _ es-> let p=(foldr maxi (-1) (map (depthV v) es)) in if (p>=0) then(1+p) else p
  _                 -> (-1)



---------------------------------------------------------------------------------------
-- isCons, devolves True if the expresion is an instance of  T(C,X) 
--         or Literal                                              --gad16ago06
---------------------------------------------------------------------------------------

isCons :: Expr -> Bool
isCons t =
   case t of
     Lit _              -> True                                     --gad16ago06
     Var _              -> True
     Comb ConsCall _ es -> True && (foldr (&&) True (map isCons es))
     _                  -> False
     
